import graphlib
from typing import Optional, Tuple, Union

from .data import RuleData, context
from .exceptions import GraphRuntimeError, InvalidPlanError
from .plan import PlanMode, Plan


class RuleEngine:
    """ Run a set of extract/transform/load rules over a dataframe.

    Takes in a plan with the definition of the extract/transform/load rules and it
    runs it over a RuleData instance. The RuleData instance can be optionally pre-populated with
    a input dataframe (in pipeline mode) or a sequence of named inputs (named dataframes).
    
    The plan can have rules to extract data (ie add more dataframes to the RuleData). It can have
    transform rules which will transform the existing dataframes (either in-place or produce new
    named dataframes). It can also have rules to load data into external systems, e.g. files,
    databases, API connections, etc.

    At the end of a plan run, the RuleData instance passed in will contain the results of the run
    (ie new dataframes/transformed dataframes) which can be inspected/operated on outside of the
    rule engine.
    """

    def __init__(self, plan: Plan):
        assert isinstance(plan, Plan)
        self.plan = plan

    def _get_context(self, data: RuleData) -> dict[str, Union[str, int, float, bool]]:
        context = {}
        context.update(self.plan.get_context())
        context.update(data.get_context())
        return context

    def run_pipeline(self, data: RuleData) -> RuleData:
        with context.set(self._get_context(data)):
            for rule in self.plan:
                rule.apply(data)
        return data

    def _get_topological_sorter(self, data: RuleData) -> graphlib.TopologicalSorter:
        g = graphlib.TopologicalSorter()
        existing_named_outputs = set(name for name, _ in data.get_named_outputs())
        named_outputs = {}
        for idx, rule in enumerate(self.plan):
            if rule.has_output():
                named_outputs_lst = list(rule.get_all_named_outputs())
                if not named_outputs_lst:
                    raise InvalidPlanError(f"Rule {rule.__class__}/(name={rule.get_name()}, index={idx}) has no named outputs.")
                for named_output in named_outputs_lst:
                    if named_output is None:
                        raise InvalidPlanError(f"Rule {rule.__class__}/(name={rule.get_name()}, index={idx}) has empty named output.")
                    existing_rule = named_outputs.get(named_output)
                    if existing_rule is not None:  
                        raise InvalidPlanError(f"Named output '{named_output}' is produced by multiple rules: {rule.__class__}/(name={rule.get_name()}) and {existing_rule[1].__class__}/(name={existing_rule[1].get_name()})")
                    named_outputs[named_output] = (idx, rule)
        named_output_clashes = existing_named_outputs & set(named_outputs.keys())
        if named_output_clashes:
            raise GraphRuntimeError(f"Named output clashes. The following named outputs are produced by rules in the plan but they also exist in the input data, leading to ambiguity: {named_output_clashes}")
        for idx, rule in enumerate(self.plan):
            if rule.has_input():
                named_inputs = list(rule.get_all_named_inputs())
                if not named_inputs:
                    raise InvalidPlanError(f"Rule {rule.__class__}/(name={rule.get_name()}, index={idx}) has no named inputs.")
                for named_input in named_inputs:
                    if named_input is None:
                        raise InvalidPlanError(f"Rule {rule.__class__}/(name={rule.get_name()}, index={idx}) has empty named input.")
                    if named_input in named_outputs:
                        g.add(idx, named_outputs[named_input][0])
                    elif named_input not in existing_named_outputs:
                        raise GraphRuntimeError(f"Rule {rule.__class__}/(name={rule.get_name()}, index={idx}) requires a named_input={named_input} which doesn't exist in the input data and it's not produced as a named output by any of the rules in the graph.")
                    else:
                        g.add(idx)
            else:
                g.add(idx)
        return g

    def run_graph(self, data: RuleData) -> RuleData:
        g = self._get_topological_sorter(data)
        g.prepare()
        with context.set(self._get_context(data)):
            while g.is_active():
                for rule_idx in g.get_ready():
                    rule = self.plan.get_rule(rule_idx)
                    rule.apply(data)
                    g.done(rule_idx)
        return data

    def validate_pipeline(self, data: RuleData) -> Tuple[bool, Optional[str]]:
        return True, None

    def validate_graph(self, data: RuleData) -> Tuple[bool, Optional[str]]:
        try:
            self._get_topological_sorter(data)
        except (InvalidPlanError, GraphRuntimeError) as exc:
            return False, str(exc)
        return True, None

    def validate(self, data: RuleData) -> Tuple[bool, Optional[str]]:
        assert isinstance(data, RuleData)
        if self.plan.is_empty():
            return False, "An empty plan cannot be run."
        mode = self.plan.get_mode()
        if mode == PlanMode.PIPELINE:
            return self.validate_pipeline(data)
        elif mode == PlanMode.GRAPH:
            return self.validate_graph(data)
        return False, "Plan's mode cannot be determined."

    def run(self, data: RuleData) -> RuleData:
        assert isinstance(data, RuleData)
        if self.plan.is_empty():
            raise InvalidPlanError("An empty plan cannot be run.")
        mode = self.plan.get_mode()
        if mode == PlanMode.PIPELINE:
            return self.run_pipeline(data)
        elif mode == PlanMode.GRAPH:
            return self.run_graph(data)
        else:
            raise InvalidPlanError("Plan's mode cannot be determined.")





class MissingColumnError(Exception):
    """ An operation is being applied to a column that is not present in the input data frame. """


class UnsupportedTypeError(Exception):
    """ A type conversion is attempted to a type that is not supported. """


class ColumnAlreadyExistsError(Exception):
    """ An attempt to create a column that already exists in the dataframe. """


class ExpressionSyntaxError(SyntaxError):
    """ A Python expression used to create a column, aggregate or other operations has a syntax error. """


class SchemaError(Exception):
    """ An operation needs a certain schema for the dataframe which is not present. """


class InvalidPlanError(Exception):
    """ The plan failed validation. """


class GraphRuntimeError(RuntimeError):
    """ There was an error when running a graph-mode plan. """


class SQLError(RuntimeError):
    """ There was an error during the execution of a sql statement. """



import yaml
from typing import Literal, Mapping, Optional, Sequence, Union

from .exceptions import InvalidPlanError
from .rule import BaseRule


class PlanMode:
    PIPELINE = "pipeline"
    GRAPH = "graph"


def plan_mode_from_rule(rule: BaseRule) -> Optional[Literal['pipeline', 'graph']]:
    if rule and rule.has_output():
        return PlanMode.GRAPH if rule.has_named_output() else PlanMode.PIPELINE


def plan_mode_from_rules(rules: Sequence[BaseRule]) -> Optional[Literal['pipeline', 'graph']]:
    for rule in rules:
        mode = plan_mode_from_rule(rule)
        if mode is not None:
            return mode


class Plan:
    """ A plan to manipulate one or multiple dataframes with a set of rules.

    A plan is a blueprint on how to extract one or more dataframes from various sources (e.g. files or
    other data sources), how to transform those dataframes by adding calculated columns, joining
    different dataframe, aggregating, sorting, etc. and ultimately how to load that into a data store
    (files or other data stores).

    A plan can operate in two modes: pipeline or graph. A pipeline graph is a simple type of plan where
    each rule take its input from the previous rule's output. A graph plan is more complex as it allows
    rules to produce named outputs which can then be used by other rules. This ultimately builds a dag
    (directed acyclic graph) of rule dependencies. A graph allows branching and joining back allowing
    complex logic. Rules are executed in the order of dependency and not in the order they are added to
    the plan. By comparison, pipelines implement a single input/single output mode where rules are
    executed in the order they are added to the plan.

    Pipeline example::

        plan = Plan()
        plan.add_rule(SortRule(['A']))
        plan.add_rule(ProjectRule(['A', 'B']))
        plan.add_rule(RenameRule({'A': 'AA', 'B': 'BB'}))
    
    Graph example::

        plan = Plan()
        plan.add_rule(SortRule(['A'], named_input="input", named_output="sorted_data"))
        plan.add_rule(ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
        plan.add_rule(RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="renamed_data"))

    Note:
        Rules that are used in graph mode should take a named_input and produce a named_output. Rules
        that use the pipeline mode must not used named inputs/outputs. The two type of rules cannot be
        used in the same plan as that leads to ambiguity.

    Args:
        mode: One of pipeline or graph, the type of the graph. Optional.
            In pipeline mode, rules don't use named inputs/outputs and they are run in the same order they are
            added to the plan, with each rule taking the input from the previous rule.
            In graph mode, rules use named inputs/outputs which create a directed acyclical graph of
            dependency. The rules are run in the order of dependency.

            When not specified, it is inferred from the first rule in the plan.
        name: A name for the plan. Optional.
        description: An optional documentation for the plan.
            This can include what the plan does, its purpose and detailed information about how it works.
        context: An optional key-value mapping which can be used in rules via string substitutions.
            It can be used as arguments into the plan to tweak the running of the plan by providing different
            values for certain arguments with each run.
            The types of the values can be: strings, int, float, boolean (True or False).
        strict: A hint about how the plan should be executed.
            When None, then the plan has no hint to provide and its the caller deciding whether to run it
            in a strict mode or not.

    Raises:
        InvalidPlanError: if pipeline mode rules are mixed with graph mode rules
    """

    def __init__(
        self,
        mode: Optional[Literal['pipeline', 'graph']]=None,
        name: Optional[str]=None,
        description: Optional[str]=None,
        context: Optional[Mapping[str, Union[str, int, float, bool]]]=None,
        strict: Optional[bool]=None
    ):
        self.mode = mode
        self.name = name
        self.description = description
        self.context = {k: v for k, v in context.items()} if context is not None else {}
        self.strict = strict
        self.rules = []

    def _check_plan_mode(self, rule: BaseRule):
        mode = self.get_mode()
        if mode is not None:
            _new_rule_mode = plan_mode_from_rule(rule)
            if _new_rule_mode is not None and mode != _new_rule_mode:
                raise InvalidPlanError(f"Mixing of rules taking named inputs and rules with no named inputs is not supported. ({mode} vs. {rule.__class__}'s mode {_new_rule_mode})")

    def get_mode(self) -> Optional[Literal['pipeline', 'graph']]:
        """ Return the mode (pipeline or graph) of the plan. """
        if self.mode is None:
            self.mode = plan_mode_from_rules(self.rules)
        return self.mode

    def get_context(self) -> dict[str, Union[str, int, float, bool]]:
        return self.context

    def add_rule(self, rule: BaseRule) -> None:
        """ Add a new rule to the plan.

        Args:
            rule: A rule instance to add to the plan
        
        Raises:
            InvalidPlanError: if the rules are mixed (pipeline vs. graph - ie. mixing use of named inputs/outputs and not using them)
        """
        assert isinstance(rule, BaseRule)
        self._check_plan_mode(rule)
        self.rules.append(rule)

    def __iter__(self):
        yield from self.rules

    def get_rule(self, idx: int) -> BaseRule:
        """ Return the rule at a certain index as per order of addition to the plan. """
        return self.rules[idx]

    def is_empty(self) -> bool:
        """ Return True if the plan has no rules, False otherwise.
        
        Returns:
            A boolean to indicate if the plan is empty.
        """
        return not self.rules

    def to_dict(self) -> dict:
        """ Serialize the plan to a dict.
        
        Returns:
            A dictionary with the plan representation.
        """
        rules = [rule.to_dict() for rule in self.rules]
        return {
            "name": self.name,
            "description": self.description,
            "context": self.context,
            "strict": self.strict,
            "rules": rules
        }

    @classmethod
    def from_dict(cls, dct: dict, backend: str, additional_packages: Optional[Sequence[str]]=None) -> 'Plan':
        """ Creates a plan instance from a python dictionary.

        Args:
            dct: A dictionary to create the plan from
            backend: One of the supported backends (ie pandas)
            additional_packages: Optional list of other packages to look for rules in
        Returns:
            A new instance of a Plan.
        """
        instance = Plan(
            name=dct.get("name"),
            description=dct.get("description"),
            context=dct.get("context"),
            strict=dct.get("strict"),
        )
        rules = dct.get("rules", ())
        for rule in rules:
            instance.add_rule(BaseRule.from_dict(rule, backend, additional_packages))
        return instance

    def to_yaml(self) -> str:
        """ Serialize the plan to yaml. """
        return yaml.safe_dump(self.to_dict())

    @classmethod
    def from_yaml(cls, yml: str, backend: str, additional_packages: Optional[Sequence[str]]=None) -> 'Plan':
        """ Creates a plan from a yaml definition.

        Args:
            yml: The yaml string to create the plan from
            backend: A supported backend (ie pandas)
            additional_packages: Optional list of other packages to look for rules in
        
        Returns:
            A new instance of a Plan.
        """
        dct = yaml.safe_load(yml)
        return cls.from_dict(dct, backend, additional_packages)

    def __eq__(self, other: 'Plan') -> bool:
        return (
            type(self) == type(other) and 
            self.name == other.name and self.description == other.description and
            self.strict == other.strict and self.rules == other.rules
        )



import importlib
import yaml
from typing import Generator, Optional, Sequence

from etlrules.data import RuleData
from etlrules.exceptions import ColumnAlreadyExistsError, MissingColumnError


class BaseRule:
    """ The base class for all rules.

    Derive your custom rules from BaseRule in order to use them in a plan.
    Implement the following methods as needed:
    apply: mandatory, it implements the functionality of the rule
    has_input: defaults to True, override and return False if your rule reads data
        into the plan and therefore has no other dataframe input
    has_output: defaults to True, override and return False if your rule writes data
        to a persistent repository and therefore has no dataframe output
    get_all_named_inputs: override to return the named inputs (if any) as strings
    get_all_named_outputs: override in case of multiple named outputs and return them as strings

    named_output (Optional[str]): Give the output of this rule a name so it can be used by another rule as a named input. Optional.
        When not set, the result of this rule will be available as the main output.
        When set to a name (string), the result will be available as that named output.
    name (Optional[str]): Give the rule a name. Optional.
        Named rules are more descriptive as to what they're trying to do/the intent.
    description (Optional[str]): Describe in detail what the rules does, how it does it. Optional.
        Together with the name, the description acts as the documentation of the rule.
    strict (bool): When set to True, the rule does a stricter valiation. Default: True

    Note:
        Add any class data members to the following list/tuples if needed:
        EXCLUDE_FROM_COMPARE: Used in implementing equality between rules. Equality is
            mostly used in tests. By default, equality looks at all data members in the
            class' __dict__. You can exclude calculated or transient data members which
            should be excluded from equality. Alternatively, you can implement __eq__ in
            your own class and not rely on the __eq__ implementation in the base class.
        EXCLUDE_FROM_SERIALIZE: Used to exclude data members from the serialization to
            dict and yaml. The serialization is implemented generically in the base class
            to serialize all data members in the class' __dict__ which do not start with
            an underscore. See the note on serialization below.
    
    Note:
        When implementing serialization, the arguments into your class should be saved as
        they are in data members with the same name as the arguments. This is because the
        de-serialization passes those as args into the __init__. As such, make sure to use
        the same names and to exclude data members which are not in the __init__ from
        serialization by adding them to EXCLUDE_FROM_SERIALIZE.

    """

    EXCLUDE_FROM_COMPARE = ()
    EXCLUDE_FROM_SERIALIZE = ()

    def __init__(self, named_output: Optional[str]=None, name: Optional[str]=None, description: Optional[str]=None, strict: bool=True):
        assert named_output is None or isinstance(named_output, str) and named_output
        self.named_output = named_output
        self.name = name
        self.description = description
        self.strict = strict

    def get_name(self) -> Optional[str]:
        """ Returns the name of the rule.

        The name is optional and it can be None.

        The name of the rule should indicate what the rule does and not how it's
        implemented. The names should read like documentation. As such, names like

        Remove duplicate first names from the list of addresses
        Only keep the names starting with A

        are preferable names to:

        DedupeRule
        ProjectRule

        Names are not used internally for anything other than your own (and your 
        end users') documentation, so use what makes sense.
        """
        return self.name

    def get_description(self) -> Optional[str]:
        """ A long description of what the rule does, why and optionally how it does it.

        The description is optional and it can be None.

        Similar to name, this long description acts as documentation for you and your users.
        It's particularly useful if your rule is serialized in a readable format like yaml
        and your users either do not have access to the documentation or they are not technical.

        Unlike the name, which should generally be a single line headline, the description is a
        long, multi-line description of the rule: the what, why, how of the rule.
        """
        return self.description

    def has_input(self) -> bool:
        """ Returns True if the rule needs a dataframe input to operate on, False otherwise.

        By default, it returns True. It should be overriden to return False for those
        rules which read data into the plan. For example, reading a csv file or reading a
        table from the DB. These are operation which do not need an input dataframe to
        operate on as they are sourcing data.
        """
        return True

    def has_output(self) -> bool:
        """ Returns True if the rule produces a dataframe, False otherwise.

        By default, it returns True. It should be overriden to return False for those
        rules which write data out of the plan. For example, writing a file or data into a
        database. These are operations which do not produce an output dataframe into
        the plan as they are writing data outside the plan.
        """
        return True

    def has_named_output(self) -> bool:
        return bool(self.named_output)

    def get_all_named_inputs(self) -> Generator[str, None, None]:
        """ Yields all the named inputs of this rule (as strings).

        By default, it yields nothing as this base rule doesn't store
        information about inputs. Some rules take no input, some take
        one or more inputs. Yield accordingly when you override.
        """
        yield from ()

    def get_all_named_outputs(self) -> Generator[str, None, None]:
        """ Yields all the named outputs of this rule (as strings).

        By default, it yields the single named_output passed into this
        rule as an argument. Some rules produce no output, some produce
        multiple outputs. Yield accordingly when you override.
        """
        yield self.named_output

    def _set_output_df(self, data, df):
        if self.named_output is None:
            data.set_main_output(df)
        else:
            data.set_named_output(self.named_output, df)

    def apply(self, data: RuleData) -> None:
        """ Applies the main rule logic to the input data.

        This is the main rule that applies a rule logic to an input data.
        The input data is an instance of RuleData which can store a single, unnamed
        dataframe (in pipeline mode) or one or many named dataframes (in graph mode).
        The rule extracts the data it needs from the data, applies its main logic
        and updates the same instance of RuleData with the output, if any.

        This method doesn't do anything in the base class other than asserting that
        the data passed in is an instance of RuleData. Override this when you derive
        from BaseRule and implement the logic of your rule.

        """
        assert isinstance(data, RuleData)

    def to_dict(self) -> dict:
        """ Serializes this rule to a python dictionary.

        This is a generic implementation that should work for all derived
        classes and therefore you shouldn't need to override, although you can do so.

        Because it aims to be generic and work correctly for all the derived classes,
        a few assumptions are made and must be respected when you implement your own
        rules derived from BaseRule.

        The class will serialize all the data attributes of a class which do not start with
        underscore and are not explicitly listed in the EXCLUDE_FROM_SERIALIZE static member
        of the class. As such, to exclude any of your internal data attributes, either named
        them so they start with an underscore or add them explicitly to EXCLUDE_FROM_SERIALIZE.

        The serialize will look into a classes __dict__ and therefore the class must have a
        __dict__.

        For the de-serialization to work generically, the name of the attributes must match the
        names of the arguments in the __init__. This is quite an important and restrictive
        constraint which is needed to avoid forcing every rule to implement a serialize/deserialize.
        
        Note:
            Use the same name for attributes on self as the respective arguments in __init__.

        """
        dct = {
            "name": self.name,
            "description": self.description,
        }
        dct.update({
            attr: value for attr, value in self.__dict__.items() 
                if not attr.startswith("_") and attr not in self.EXCLUDE_FROM_SERIALIZE
                and attr not in dct.keys()
        })
        return {
            self.__class__.__name__: dct
        }

    @classmethod
    def from_dict(cls, dct: dict, backend: str, additional_packages: Optional[Sequence[str]]=None) -> 'BaseRule':
        """ Creates a rule instance from a python dictionary.

        Args:
            dct: A dictionary to create the plan from
            backend: One of the supported backends (ie pandas)
            additional_packages: Optional list of other packages to look for rules in
        Returns:
            A new instance of a Plan.
        """
        assert backend and isinstance(backend, str)
        keys = tuple(dct.keys())
        assert len(keys) == 1
        rule_name = keys[0]
        backend_pkgs = [f'etlrules.backends.{backend}']
        for additional_package in additional_packages or ():
            backend_pkgs.append(additional_package)
        modules = [importlib.import_module(backend_pkg, '') for backend_pkg in backend_pkgs]
        for mod in modules:
            clss = getattr(mod, rule_name, None)
            if clss is not None:
                break
        assert clss, f"Cannot find class {rule_name} in packages: {backend_pkgs}"
        if clss is not cls:
            return clss.from_dict(dct, backend, additional_packages)
        return clss(**dct[rule_name])

    def to_yaml(self):
        """ Serialize the rule to yaml. """
        return yaml.safe_dump(self.to_dict())

    @classmethod
    def from_yaml(cls, yml: str, backend: str, additional_packages: Optional[Sequence[str]]=None) -> 'BaseRule':
        """ Creates a rule instance from a yaml definition.

        Args:
            yml: The yaml string to create the plan from
            backend: A supported backend (ie pandas)
            additional_packages: Optional list of other packages to look for rules in
        
        Returns:
            A new instance of a rule.
        """
        dct = yaml.safe_load(yml)
        return cls.from_dict(dct, backend, additional_packages)

    def __eq__(self, other) -> bool:
        return (
            type(self) == type(other) and 
            {k: v for k, v in self.__dict__.items() if k not in self.EXCLUDE_FROM_COMPARE} == 
            {k: v for k, v in other.__dict__.items() if k not in self.EXCLUDE_FROM_COMPARE}
        )


class UnaryOpBaseRule(BaseRule):
    """ Base class for unary operation rules (ie operations taking a single data frame as input). """

    def __init__(self, named_input: Optional[str]=None, named_output: Optional[str]=None, name: Optional[str]=None, description: Optional[str]=None, strict: bool=True):
        super().__init__(named_output=named_output, name=name, description=description, strict=strict)
        assert named_input is None or isinstance(named_input, str) and named_input
        self.named_input = named_input

    def _get_input_df(self, data: RuleData):
        if self.named_input is None:
            return data.get_main_output()
        return data.get_named_output(self.named_input)

    def get_all_named_inputs(self):
        yield self.named_input


class ColumnsInOutMixin:
    def validate_input_column(self, df_columns: Sequence[str], input_column: str, strict: bool):
        if input_column not in df_columns:
            raise MissingColumnError(f"Column '{input_column}' is missing from the input dataframe.")
        return input_column

    def validate_output_column(self, df_columns: Sequence[str], input_column: str, output_column: Optional[str], strict: bool):
        if output_column is not None:
            if strict and output_column in df_columns:
                raise ColumnAlreadyExistsError(f"Column '{output_column}' already exists in the input dataframe.")
            return output_column
        return input_column

    def validate_in_out_columns(self, df_columns: Sequence[str], input_column: str, output_column: Optional[str], strict: bool):
        input_column = self.validate_input_column(df_columns, input_column, strict)
        output_column = self.validate_output_column(df_columns, input_column, output_column, strict)
        return input_column, output_column

    def validate_columns_in(self, df_columns: Sequence[str], columns: Sequence[str], strict: bool) -> Sequence[str]:
        if not set(columns) <= set(df_columns):
            raise MissingColumnError(f"Column(s) {set(columns) - set(df_columns)} are missing from the input dataframe.")
        return columns

    def validate_columns_out(self, df_columns: Sequence[str], columns: Sequence[str], output_columns: Optional[Sequence[str]], strict: bool, validate_length: bool=True) -> Sequence[str]:
        if output_columns:
            if strict:
                existing_columns = set(output_columns) & set(df_columns)
                if existing_columns:
                    raise ColumnAlreadyExistsError(f"Column(s) already exist: {existing_columns}")
            if validate_length and len(output_columns) != len(columns):
                raise ValueError(f"output_columns must be of the same length as the columns: {columns}")
        else:
            output_columns = columns
        return output_columns

    def validate_columns_in_out(self, df_columns: Sequence[str], columns: Sequence[str], output_columns: Optional[Sequence[str]], strict: bool, validate_length: bool=True) -> tuple[Sequence[str], Sequence[str]]:
        columns = self.validate_columns_in(df_columns, columns, strict)
        output_columns = self.validate_columns_out(df_columns, columns, output_columns, strict, validate_length=validate_length)
        return columns, output_columns


class BinaryOpBaseRule(BaseRule):
    """ Base class for binary operation rules (ie operations taking two data frames as input). """

    def __init__(self, named_input_left: Optional[str], named_input_right: Optional[str], named_output: Optional[str]=None, name: Optional[str]=None, description: Optional[str]=None, strict: bool=True):
        super().__init__(named_output=named_output, name=name, description=description, strict=strict)
        assert named_input_left is None or isinstance(named_input_left, str) and named_input_left
        assert named_input_right is None or isinstance(named_input_right, str) and named_input_right
        self.named_input_left = named_input_left
        self.named_input_right = named_input_right

    def _get_input_df_left(self, data: RuleData):
        if self.named_input_left is None:
            return data.get_main_output()
        return data.get_named_output(self.named_input_left)

    def _get_input_df_right(self, data: RuleData):
        if self.named_input_right is None:
            return data.get_main_output()
        return data.get_named_output(self.named_input_right)

    def get_all_named_inputs(self):
        yield self.named_input_left
        yield self.named_input_right



import pytest

from etlrules.data import context


def test_no_context():
    with pytest.raises(RuntimeError) as exc:
        context.KEY
    assert exc.value.args[0] == "No context set."


def test_invalid_type():
    with pytest.raises(TypeError) as exc:
        context[0]
    assert exc.value.args[0] == "Context attr name must be a string."


def test_context():
    with context.set({"STR": "key", "INT": 1, "FLOAT": 2.5, "BOOL": True}) as ctx:
        assert context.STR == "key"
        assert context.INT == 1
        assert context.FLOAT == 2.5
        assert context.BOOL == True
        assert context["STR"] == "key"
        assert context["INT"] == 1
        assert context["FLOAT"] == 2.5
        assert context["BOOL"] == 1
        assert ctx["STR"] == "key"
        assert ctx["INT"] == 1
        assert ctx["FLOAT"] == 2.5
        assert ctx["BOOL"] == 1
        with pytest.raises(KeyError) as exc:
            context.KEY2
        assert exc.value.args[0] == "No such attribute 'KEY2' found in the current context."
    with pytest.raises(RuntimeError) as exc:
        context.KEY
    assert exc.value.args[0] == "No context set."


import argparse
import logging
import os
import shutil
import tempfile
from typing import Any, Optional

from .data import RuleData
from .engine import RuleEngine
from .plan import Plan


logger = logging.getLogger(__name__)


def get_args_parser(plan: Optional[Plan]=None) -> dict[str, Any]:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--plan",
        help='Specify a yaml file containing a plan to run.',
        required=True,
    )
    parser.add_argument(
        "-b",
        "--backend",
        help="The backend to use for running the plan.",
        choices=["pandas", "polars", "dask"],
        required=False,
        default="pandas"
    )
    if plan:
        context = plan.get_context()
        for key, val in context.items():
            if isinstance(val, int):
                val_type = int
            elif isinstance(val, float):
                val_type = float
            else:
                val_type = str
            parser.add_argument(
                "--" + key,
                required=False,
                default=val,
                type=val_type
            )
        args = parser.parse_args()
    else:
        args, _ = parser.parse_known_args()
    return vars(args)


def load_plan(plan_file: str, backend: str) -> Plan:
    """ Load a plan from a yaml file.

    Basic usage:

        from etlrules import load_plan
        plan = load_plan("/home/someuser/some_plan.yml", "pandas")

    Args:
        plan_file: A path to a yaml file with the plan definition
        backend: One of the supported backends (e.g. pandas, polars, etc.)

    Returns:
        A Plan instance deserialized from the yaml file.
    """
    with open(plan_file, 'rt') as plan_f:
        contents = plan_f.read()
    return Plan.from_yaml(contents, backend)


def get_etlrules_temp_dir() -> tuple[str, bool]:
    etlrules_tempdir_cleanup = True
    etlrules_tempdir = os.environ.get("ETLRULES_TEMPDIR")
    if etlrules_tempdir:
        if os.path.exists(etlrules_tempdir) and os.path.isdir(etlrules_tempdir):
            etlrules_tempdir_cleanup = False
        else:
            os.makedirs(etlrules_tempdir)
    else:
        etlrules_tempdir = tempfile.mkdtemp(prefix='tmp_etlrules')
    return etlrules_tempdir, etlrules_tempdir_cleanup


def run_plan(plan_file: str, backend: str) -> RuleData:
    """ Runs a plan from a yaml file with a given backend.

    The backend referers to the underlying dataframe library used to run
    the plan.

    Basic usage:

        from etlrules import run_plan
        data = run_plan("/home/someuser/some_plan.yml", "pandas")

    Args:
        plan_file: A path to a yaml file with the plan definition
        backend: One of the supported backends

    Note:
        The supported backends:
            pandas, polars, dask (work in progress)

    Returns:
        A RuleData instance which contains the result dataframe(s).
    """
    plan = load_plan(plan_file, backend)
    args = get_args_parser(plan)
    context = {}
    context.update(args)
    etlrules_tempdir, etlrules_tempdir_cleanup = get_etlrules_temp_dir()
    context.update({
        "etlrules_tempdir": etlrules_tempdir,
        "etlrules_tempdir_cleanup": etlrules_tempdir_cleanup,
    })
    try:
        data = RuleData(context=context)
        engine = RuleEngine(plan)
        engine.run(data)
    finally:
        if etlrules_tempdir_cleanup:
            shutil.rmtree(etlrules_tempdir)

    return data


def run() -> None:
    args = get_args_parser()
    logger.info(f"Running plan '{args['plan']}' with backend: {args['backend']}")
    run_plan(args["plan"], args["backend"])
    logger.info("Done.")


if __name__ == "__main__":
    run()



"""Top-level package for ETLrules."""

from .data import RuleData, context
from .engine import RuleEngine
from .exceptions import (
    ColumnAlreadyExistsError, ExpressionSyntaxError, GraphRuntimeError,
    InvalidPlanError, MissingColumnError, SchemaError, UnsupportedTypeError,
)
from .plan import Plan, PlanMode
from .runner import load_plan, run_plan


__author__ = """Ciprian Miclaus"""
__email__ = "ciprianm@gmail.com"
__version__ = "0.3.2"


__all__ = [
    "RuleData", "context",
    "RuleEngine",
    "ColumnAlreadyExistsError", "ExpressionSyntaxError", "GraphRuntimeError",
    "InvalidPlanError", "MissingColumnError", "SchemaError", "UnsupportedTypeError",
    "Plan", "PlanMode",
    "load_plan", "run_plan",
]



from contextlib import contextmanager
from typing import Generator, Mapping, Optional, Union


class RuleData:
    def __init__(self,
        main_input=None,
        named_inputs=None,
        context: Optional[Mapping[str, Union[str, int, float, bool]]]=None,
        strict: bool=True
    ):
        self.strict = strict
        self.main_output = main_input
        self.named_outputs = (
            {name: df for name, df in named_inputs.items()} if named_inputs else {}
        )
        self.context = {k: v for k, v in context.items()} if context is not None else {}
        self.lineage_info = {}

    def get_main_output(self):
        return self.main_output

    def set_main_output(self, df):
        self.main_output = df

    def get_named_output(self, name: str):
        assert name in self.named_outputs, f"No such named output {name}"
        return self.named_outputs[name]

    def set_named_output(self, name, df):
        if self.strict:
            assert (
                name not in self.named_outputs
            ), f"{name} already exists as a named output. It will be overwritten."
        self.named_outputs[name] = df

    def get_named_outputs(self):
        yield from self.named_outputs.items()

    def get_context(self) -> dict[str, Union[str, int, float, bool]]:
        return self.context


class Context:

    def __init__(self):
        self.mappers = []

    @contextmanager
    def set(self, mapping: Mapping[str, Union[str, int, float, bool]]) -> Generator[dict[str, str], None, None]:
        self.mappers.append(
            {k: v for k, v in mapping.items()}
        )
        try:
            yield self.mappers[-1]
        finally:
            self.mappers.pop()

    def _do_get_attr(self, attr_name: str) -> Union[str, int, float, bool]:
        if not isinstance(attr_name, str):
            raise TypeError("Context attr name must be a string.")
        if not self.mappers:
            raise RuntimeError("No context set.")
        for current_context in reversed(self.mappers):
            try:
                return current_context[attr_name]
            except KeyError:
                ...
        raise KeyError(f"No such attribute '{attr_name}' found in the current context.")


    def __getattr__(self, attr_name: str) -> Union[str, int, float, bool]:
        return self._do_get_attr(attr_name)

    def __getitem__(self, attr_name: str) -> Union[str, int, float, bool]:
        return self._do_get_attr(attr_name)


context = Context()



import os
from pathlib import Path
import pytest
import sys
from unittest.mock import patch

from etlrules.runner import run
from etlrules.backends.common.io.db import SQLAlchemyEngines


EXPECTED = [
    (1, 'Mike', 'Geoffrey', 'Like a python', 2012),
    (2, 'Adam', 'Rolley', 'How to make friends', 1998),
    (3, 'Michelle', 'Saville', 'Scent of a pear', 2022),
    (4, 'Dorothy', 'Andrews', 'Tell me your name', 2014),
    (5, 'John', 'Rawley', 'Make me a lasagna', 2011),
]


@pytest.mark.parametrize("backend_str", [
    "pandas",
    "polars",
])
def test_runner(backend_str):
    args = ["runner.py", "-p", "./tests/csv2db.yml", "-b", backend_str]
    with patch.object(sys, 'argv', args):
        run()
    try:
        import sqlalchemy as sa
        engine = SQLAlchemyEngines.get_engine("sqlite:///tests/mydb.db")
        with engine.connect() as connection:
            rows = connection.execute(sa.text("SELECT * FROM MyTable")).fetchall()
            assert rows == EXPECTED
    finally:
        os.remove(Path("tests") / "mydb.db")


@pytest.mark.parametrize("backend_str", [
    "pandas",
    "polars",
])
def test_runner_with_cli_overrides(backend_str):
    db_name = "otherdb.db"
    args = [
        "runner.py", "-p", "./tests/csv2db.yml", "-b", backend_str,
        "--csv_file_dir", "./tests",
        "--csv_file_name", "csv_sample.csv",
        "--sql_engine", f"sqlite:///tests/{db_name}",
        "--sql_table", "SomeOtherTable"
    ]
    with patch.object(sys, 'argv', args):
        run()
    try:
        import sqlalchemy as sa
        engine = SQLAlchemyEngines.get_engine(f"sqlite:///tests/{db_name}")
        with engine.connect() as connection:
            rows = connection.execute(sa.text("SELECT * FROM SomeOtherTable")).fetchall()
            assert rows == EXPECTED
    finally:
        os.remove(Path("tests") / db_name)



import pytest

from etlrules.data import RuleData
from etlrules.engine import RuleEngine
from etlrules.exceptions import GraphRuntimeError, InvalidPlanError
from etlrules.plan import Plan

from tests.utils.data import assert_frame_equal


def test_run_simple_plan(backend):
    input_df = backend.DataFrame(data=[
        {'A': 2, 'B': 'n', 'C': True},
        {'A': 1, 'B': 'm', 'C': False},
        {'A': 3, 'B': 'p', 'C': True},
    ])
    data = RuleData(input_df)
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A']))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B']))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}))
    rule_engine = RuleEngine(plan)
    valid, err = rule_engine.validate(data)
    assert valid is True
    assert err is None
    rule_engine.run(data)
    result = data.get_main_output()
    expected = backend.DataFrame(data=[
        {'AA': 1, 'BB': 'm'},
        {'AA': 2, 'BB': 'n'},
        {'AA': 3, 'BB': 'p'},
    ])
    assert_frame_equal(result, expected)


def test_run_simple_plan_named_inputs(backend):
    input_df = backend.DataFrame(data=[
        {'A': 2, 'B': 'n', 'C': True},
        {'A': 1, 'B': 'm', 'C': False},
        {'A': 3, 'B': 'p', 'C': True},
    ])
    data = RuleData(named_inputs={"input": input_df})
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="renamed_data"))
    rule_engine = RuleEngine(plan)
    valid, err = rule_engine.validate(data)
    assert valid is True
    assert err is None
    rule_engine.run(data)
    result = data.get_named_output("renamed_data")
    expected = backend.DataFrame(data=[
        {'AA': 1, 'BB': 'm'},
        {'AA': 2, 'BB': 'n'},
        {'AA': 3, 'BB': 'p'},
    ])
    assert_frame_equal(result, expected)


def test_mix_pipeline_graph_plan_types(backend):
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A']))
    with pytest.raises(InvalidPlanError):
        plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))


def test_mix_graph_pipeline_plan_types(backend):
    plan = Plan()
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    with pytest.raises(InvalidPlanError):
        plan.add_rule(backend.rules.SortRule(['A']))


def test_run_empty_plan(backend):
    data = RuleData(named_inputs={"input": backend.DataFrame(data=[])})
    plan = Plan()
    rule_engine = RuleEngine(plan)
    with pytest.raises(InvalidPlanError) as exc:
        rule_engine.run(data)
    assert str(exc.value) == "An empty plan cannot be run."
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "An empty plan cannot be run." in err
    assert valid is False


def test_run_unknown_mode_plan(backend):
    data = RuleData(named_inputs={"input": backend.DataFrame(data=[])})
    plan = Plan()
    plan.add_rule(backend.rules.WriteCSVFileRule(file_name="test.csv.gz", file_dir="/home/myuser", separator=",", header=True, compression="gzip",
                named_input="result", name="BF", description="Some desc2 BF", strict=True))
    rule_engine = RuleEngine(plan)
    with pytest.raises(InvalidPlanError) as exc:
        rule_engine.run(data)
    assert str(exc.value) == "Plan's mode cannot be determined."
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "Plan's mode cannot be determined." in err
    assert valid is False


def test_run_simple_plan_named_inputs_different_order(backend):
    input_df = backend.DataFrame(data=[
        {'A': 2, 'B': 'n', 'C': True},
        {'A': 1, 'B': 'm', 'C': False},
        {'A': 3, 'B': 'p', 'C': True},
    ])
    data = RuleData(named_inputs={"input": input_df})
    plan = Plan()
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="renamed_data"))
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    rule_engine = RuleEngine(plan)
    valid, err = rule_engine.validate(data)
    assert valid is True
    assert err is None
    rule_engine.run(data)
    result = data.get_named_output("renamed_data")
    expected = backend.DataFrame(data=[
        {'AA': 1, 'BB': 'm'},
        {'AA': 2, 'BB': 'n'},
        {'AA': 3, 'BB': 'p'},
    ])
    assert_frame_equal(result, expected)


def test_run_missing_named_input(backend):
    data = RuleData()
    plan = Plan()
    plan.add_rule(backend.rules.ReadCSVFileRule(file_name="test.csv.gz", file_dir="/home/myuser", named_output="input2"))
    plan.add_rule(backend.rules.WriteCSVFileRule(file_name="test.csv.gz", file_dir="/home/myuser", separator=",", header=True, compression="gzip",
                named_input="sorted_data", name="BF", description="Some desc2 BF", strict=True))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="renamed_data"))
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    rule_engine = RuleEngine(plan)
    with pytest.raises(GraphRuntimeError) as exc:
        rule_engine.run(data)
    assert "requires a named_input=input which doesn't exist in the input data and it's not produced as a named output" in str(exc.value)
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "requires a named_input=input which doesn't exist in the input data and it's not produced as a named output" in err
    assert valid is False


def test_run_missing_named_input_in_rule(backend):
    data = RuleData()
    plan = Plan()
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_output="renamed_data"))
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    rule_engine = RuleEngine(plan)
    with pytest.raises(InvalidPlanError) as exc:
        rule_engine.run(data)
    assert "RenameRule" in str(exc.value)
    assert "has empty named input." in str(exc.value)
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "RenameRule" in err
    assert "has empty named input." in err
    assert valid is False


def test_run_missing_named_output_clashes(backend):
    data = RuleData()
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B', 'C'], named_input="input", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="renamed_data"))
    rule_engine = RuleEngine(plan)
    with pytest.raises(InvalidPlanError) as exc:
        rule_engine.run(data)
    assert "Named output 'projected_data' is produced by multiple rules" in str(exc.value)
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "Named output 'projected_data' is produced by multiple rules" in err
    assert valid is False

def test_run_produces_output_already_exists_in_input_data(backend):
    data = RuleData(named_inputs={"input": backend.DataFrame(data=[])})
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data", named_output="input"))
    rule_engine = RuleEngine(plan)
    with pytest.raises(GraphRuntimeError) as exc:
        rule_engine.run(data)
    assert "Named output clashes. The following named outputs are produced by rules in the plan but they also exist in the input data, leading to ambiguity: {'input'}" in str(exc.value)
    valid, err = rule_engine.validate(data)
    assert err is not None
    assert "Named output clashes. The following named outputs are produced by rules in the plan but they also exist in the input data, leading to ambiguity: {'input'}" in err
    assert valid is False



import pytest

from etlrules.exceptions import InvalidPlanError
from etlrules.plan import Plan


def test_plan_rules(backend):
    plan = Plan(name="Test Plan", description="This is a plan to test the plan.", strict=True)
    rules_in = [
        backend.rules.SortRule(['A']),
        backend.rules.ProjectRule(['A', 'B']),
        backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}),
    ]
    for rule in rules_in:
        plan.add_rule(rule)
    for idx, rule in enumerate(plan):
        assert rule == rules_in[idx]


def test_plan_to_from_dict(backend):
    plan = Plan(name="Test Plan", description="This is a plan to test the plan.", strict=True)
    plan.add_rule(backend.rules.SortRule(['A']))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B']))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}))
    dct = plan.to_dict()
    assert dct["name"] == "Test Plan"
    assert dct["description"] == "This is a plan to test the plan."
    assert dct["strict"] == True
    assert len(dct["rules"]) == 3
    plan2 = Plan.from_dict(dct, backend.name)
    assert plan == plan2


def test_plan_to_from_yaml(backend):
    plan = Plan(name="Test Plan", description="This is a plan to test the plan.", strict=True)
    plan.add_rule(backend.rules.SortRule(['A']))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B']))
    plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}))
    yml = plan.to_yaml()
    plan2 = Plan.from_yaml(yml, backend.name)
    assert plan == plan2


def test_graph_plan_rule_without_named_output(backend):
    plan = Plan()
    plan.add_rule(backend.rules.SortRule(['A'], named_input="input", named_output="sorted_data"))
    plan.add_rule(backend.rules.ProjectRule(['A', 'B'], named_input="sorted_data", named_output="projected_data"))
    with pytest.raises(InvalidPlanError) as exc:
        plan.add_rule(backend.rules.RenameRule({'A': 'AA', 'B': 'BB'}, named_input="projected_data"))
    assert "Mixing of rules taking named inputs and rules with no named inputs is not supported." in str(exc.value)
